"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.BitbucketServerApi = void 0;
const lodash_1 = __importDefault(require("lodash"));
const request_promise_1 = __importDefault(require("request-promise"));
class BitbucketServerApi {
    constructor(inputs) {
        this.serverURL = inputs.bitbucketHost;
        this.username = inputs.bitbucketUsername;
        this.password = inputs.bitbucketPassword;
    }
    async getRepo(repo) {
        const [workspace, repoSlug] = this._getRepoSlugAndWorkspaceFromRepo(repo);
        return await this._withPersonalWorkspaceFallback((attemptedWorkspace) => {
            return this._sendGetRequest(`rest/api/latest/projects/${attemptedWorkspace}/repos/${repoSlug}`);
        }, workspace);
    }
    async getBranch(repo, branch) {
        const [workspace, repoSlug] = this._getRepoSlugAndWorkspaceFromRepo(repo);
        const result = await this._sendGetRequest(`rest/api/latest/projects/${workspace}/repos/${repoSlug}/commits?until=${branch}&limit=0&start=0`);
        const branchInfo = lodash_1.default.get(result, 'values[0]');
        if (!branchInfo) {
            return null;
        }
        branchInfo.commitURL = this._buildCommitUrl(workspace, repoSlug, branchInfo.id);
        return branchInfo;
    }
    async getPullRequests(repo, branch) {
        const [workspace, repoSlug] = this._getRepoSlugAndWorkspaceFromRepo(repo);
        const result = await this._sendGetRequest(`rest/api/latest/projects/${workspace}/repos/${repoSlug}/pull-requests?start=0&order=newest&state=ALL&at=refs/heads/${branch}&direction=outgoing`);
        return lodash_1.default.get(result, 'values', []);
    }
    async getCommitsInfo(repo, pr) {
        const [workspace, repoSlug] = this._getRepoSlugAndWorkspaceFromRepo(repo);
        let result = await this._sendGetRequest(`projects/${workspace}/repos/${repoSlug}/pull-requests/${pr.id}/commits?start=0&avatarSize=64&limit=50&contents`);
        const reversedCommits = lodash_1.default.get(result, 'values', []);
        while (!result.isLastPage) {
            result = await this._sendGetRequest(`projects/${workspace}/repos/${repoSlug}/pull-requests/${pr.id}/commits?start=${result.nextPageStart}&limit=50&avatarSize=64&contents`);
            reversedCommits.push(...result.values);
        }
        const commits = lodash_1.default.reverse(reversedCommits);
        if (commits && commits.length > 0) {
            const firstCommitDate = commits[0].committerTimestamp;
            const lastCommitDate = lodash_1.default.last(commits)?.committerTimestamp;
            const commitsByUser = {};
            const committersMap = {};
            for (const commit of commits) {
                const userName = commit.author.emailAddress;
                const avatar = commit.author.avatarUrl;
                committersMap[userName] = {
                    userName,
                    avatar,
                };
                if (!commitsByUser[userName]) {
                    commitsByUser[userName] = [];
                }
                commitsByUser[userName].push({
                    url: this._buildCommitUrl(workspace, repoSlug, commit.id),
                    userName,
                    sha: commit.id,
                    message: commit.message,
                    commitDate: commit.committerTimestamp
                });
            }
            return {
                committers: lodash_1.default.values(committersMap),
                commits: lodash_1.default.flatten(lodash_1.default.values(commitsByUser)),
                firstCommitDate,
                lastCommitDate
            };
        }
        return null;
    }
    async _sendGetRequest(uri) {
        return (0, request_promise_1.default)({
            method: 'GET',
            uri: `${this.serverURL}/${uri}`,
            auth: {
                user: this.username,
                pass: this.password
            },
            json: true
        });
    }
    _buildCommitUrl(workspace, repoSlug, commitId) {
        if (!commitId) {
            return;
        }
        return `${this.serverURL}/projects/${workspace}/repos/${repoSlug}/commits/${commitId}`;
    }
    _getRepoSlugAndWorkspaceFromRepo(repo) {
        return repo.split('/');
    }
    async _withPersonalWorkspaceFallback(cb, workspace) {
        return await cb(workspace).catch(async (error) => {
            if (error.statusCode === 404) {
                const personalWorkspace = `~${workspace}`;
                console.log(`looking for personal workspace ${personalWorkspace} as a fallback`);
                return await cb(personalWorkspace);
            }
            throw error;
        });
    }
}
exports.BitbucketServerApi = BitbucketServerApi;
//# sourceMappingURL=bitbucket-server.api.js.map